"""Wraps cbor2 with hooks for encoding and decoding tensors."""
import jax
import cbor2
import numpy as np
import functools

from cbor2 import CBORTag

# Standard tags for multidimensional arrays from RFC8746
# (little-endian, row-major).
TAG_FLOAT32 = 85
TAG_FLOAT64 = 86
TAG_INT32 = 78
TAG_INT64 = 79
TAG_ARRAY = 40

def encode_flat(arr):
    if arr.dtype == np.float32:
        return CBORTag(TAG_FLOAT32, arr.tobytes())
    if arr.dtype == np.int32:
        return CBORTag(TAG_INT32, arr.tobytes())
    else:
        raise NotImplemented

def default_encoder(encoder, value):
    if isinstance(value, jax.numpy.DeviceArray):
        encoder.encode(np.array(value))
    elif isinstance(value, np.ndarray):
        encoder.encode(CBORTag(TAG_ARRAY, [list(value.shape), encode_flat(value)]))
    else:
        raise NotImplemented

def tag_hook(decoder, tag, shareable_index=None):
    if tag.tag == TAG_ARRAY:
        [shape, value] = tag.value
        return value.reshape(shape)
    elif tag.tag == TAG_FLOAT32:
        return np.frombuffer(tag.value, dtype=np.float32)
    elif tag.tag == TAG_INT32:
        return np.frombuffer(tag.value, dtype=np.int32)
    elif tag.tag == TAG_INT64:
        return np.frombuffer(tag.value, dtype=np.int64)
    else:
        return tag

dumps = functools.partial(cbor2.dumps, default=default_encoder)
dump = functools.partial(cbor2.dump, default=default_encoder)

loads = functools.partial(cbor2.loads, tag_hook=tag_hook)
load = functools.partial(cbor2.load, tag_hook=tag_hook)



import jax
import jax.numpy as jnp

def upsample2x_base(image, kernel):
    ksize = kernel.shape[0]
    (n, c, h, w) = image.shape
    out = jax.lax.conv_general_dilated(image.reshape(n*c,1,h,w),
                                       kernel.reshape(1,1,ksize,ksize),
                                       window_strides=[1,1],
                                       padding=[(ksize//2,ksize//2),(ksize//2,ksize//2)],
                                       lhs_dilation=[2,2],
                                       rhs_dilation=None,
                                       dimension_numbers=('NCHW',
                                                          'IOHW', 'NCHW'))

    # normalization for parts that touch the zero-padding
    norm = jax.lax.conv_general_dilated(jnp.ones((1,1,h,w)),
                                        kernel.reshape(1,1,ksize,ksize),
                                        window_strides=[1,1],
                                        padding=[(ksize//2,ksize//2),(ksize//2,ksize//2)],
                                        lhs_dilation=[2,2],
                                        rhs_dilation=None,
                                        dimension_numbers=('NCHW',
                                                           'IOHW', 'NCHW'))
    return (out / norm).reshape(n, c, 2*h,2*w)

def upsample2x(image, method='linear'):
    if method == 'lanczos3':
        # extracted from the gradients of jax.image.resize(method='lanczos3')
        kernel = jnp.array([0.0073782638646662235, 0.030112292617559433,
                            -0.06799723953008652, -0.13327467441558838,
                            0.2710106074810028, 0.8927707076072693,
                            0.8927707672119141, 0.2710106074810028,
                            -0.13327467441558838, -0.06799724698066711,
                            0.03011229634284973, 0.007378263399004936])
    elif method == 'cubic':
        # extracted from the gradients of jax.image.resize(method='cubic')
        kernel = jnp.array([-0.0234375, -0.0703125, 0.2265625, 0.8671875, 0.8671875, 0.2265625, -0.0703125, -0.0234375])
    elif method == 'linear':
        # extracted from the gradients of jax.image.resize(method='linear')
        kernel = jnp.array([0.25, 0.75, 0.75, 0.25])
    kernel = kernel.reshape(-1,1) * kernel.reshape(1,-1)

    return upsample2x_base(image, kernel)

def downsample2x_base(image, kernel):
    ksize = kernel.shape[0]
    (n, c, h, w) = image.shape
    out = jax.lax.conv_general_dilated(image.reshape(n*c,1,h,w),
                                       kernel.reshape(1,1,ksize,ksize),
                                       window_strides=[2,2],
                                       padding=[(ksize//2-1,ksize//2-1),(ksize//2-1,ksize//2-1)],
                                       lhs_dilation=[1,1],
                                       rhs_dilation=None,
                                       dimension_numbers=('NCHW',
                                                          'IOHW', 'NCHW'))

    # normalization for parts that touch the zero-padding
    norm = jax.lax.conv_general_dilated(jnp.ones((1,1,h,w)),
                                        kernel.reshape(1,1,ksize,ksize),
                                        window_strides=[2,2],
                                        padding=[(ksize//2-1,ksize//2-1),(ksize//2-1,ksize//2-1)],
                                        lhs_dilation=[1,1],
                                        rhs_dilation=None,
                                        dimension_numbers=('NCHW',
                                                           'IOHW', 'NCHW'))
    return (out / norm).reshape(n, c, h//2,w//2)

def downsample2x(image, method='linear'):
    if method == 'linear':
        kernel = jnp.array([0.125, 0.375, 0.375, 0.125])
    elif method == 'cubic':
        kernel = jnp.array([-0.01171875, -0.03515625, 0.11328125, 0.43359375, 0.43359375, 0.11328125, -0.03515625, -0.01171875])
    elif method == 'lanczos3':
        kernel = jnp.array([0.003689131001010537, 0.015056144446134567, -0.03399861603975296,
                            -0.066637322306633, 0.13550527393817902, 0.44638532400131226,
                            0.44638532400131226, 0.13550527393817902, -0.066637322306633,
                            -0.03399861603975296, 0.015056144446134567, 0.003689131001010537])
    kernel = kernel.reshape(-1,1) * kernel.reshape(1,-1)
    return downsample2x_base(image, kernel)



import sys
sys.path = ['.'] + sys.path
import torch
import numpy as np
import jax
import jax.numpy as jnp
import jaxtorch
from jaxtorch.core import Module, Context, PRNG

torch.backends.cudnn.deterministic = True

def check(old, new, limit=1e-6):
    old_np = np.array(old)
    new_np = np.array(new)
    assert old_np.shape == new_np.shape
    difference = np.abs(old_np - new_np).max()
    magnitude = np.mean((np.abs(old_np) + np.abs(new_np))/2)
    # Pretty loose bounds due to float32 precision. There seem to be
    # some implementation differences resulting in different error
    # bounds betwen jax and torch.
    assert difference/magnitude < limit, difference/magnitude

def totorch(x):
    return torch.tensor(np.array(x))
def fromtorch(x):
    return jnp.array(np.array(x))

@torch.no_grad()
def test_conv1d():
    rng = PRNG(jax.random.PRNGKey(0))
    x = jax.random.normal(key=rng.split(), shape=[1, 4, 5])
    w = jax.random.normal(key=rng.split(), shape=[3, 4, 3])
    b = jax.random.normal(key=rng.split(), shape=[3])

    x_torch = totorch(x)
    w_torch = totorch(w)
    b_torch = totorch(b)

    new_result = jaxtorch.nn.functional.conv1d(x, w, bias=b, padding='same')
    old_result = torch.nn.functional.conv1d(x_torch, w_torch, bias=b_torch, padding='same')
    check(old_result, new_result)

    new_result = jaxtorch.nn.functional.conv1d(x, w, padding='valid')
    old_result = torch.nn.functional.conv1d(x_torch, w_torch, padding='valid')
    check(old_result, new_result)

    new_result = jaxtorch.nn.functional.conv1d(x, w, padding='valid', stride=2)
    old_result = torch.nn.functional.conv1d(x_torch, w_torch, padding='valid', stride=2)
    check(old_result, new_result)

    new_result = jaxtorch.nn.functional.conv1d(x, w, padding='valid', dilation=2)
    old_result = torch.nn.functional.conv1d(x_torch, w_torch, padding='valid', dilation=2)
    check(old_result, new_result)

    w = jax.random.normal(key=rng.split(), shape=[4, 2, 3])
    w_torch = totorch(w)
    new_result = jaxtorch.nn.functional.conv1d(x, w, padding='valid', groups=2)
    old_result = torch.nn.functional.conv1d(x_torch, w_torch, padding='valid', groups=2)
    check(old_result, new_result)

@torch.no_grad()
def test_conv2d():
    rng = PRNG(jax.random.PRNGKey(0))
    x = jax.random.normal(key=rng.split(), shape=[1, 4, 5, 5])
    w = jax.random.normal(key=rng.split(), shape=[3, 4, 3, 3])
    b = jax.random.normal(key=rng.split(), shape=[3])

    x_torch = totorch(x)
    w_torch = totorch(w)
    b_torch = totorch(b)

    new_result = jaxtorch.nn.functional.conv2d(x, w, bias=b, padding='same')
    old_result = torch.nn.functional.conv2d(x_torch, w_torch, bias=b_torch, padding='same')
    check(old_result, new_result)

    new_result = jaxtorch.nn.functional.conv2d(x, w, padding='valid')
    old_result = torch.nn.functional.conv2d(x_torch, w_torch, padding='valid')
    check(old_result, new_result)

    new_result = jaxtorch.nn.functional.conv2d(x, w, padding='valid', stride=2)
    old_result = torch.nn.functional.conv2d(x_torch, w_torch, padding='valid', stride=2)
    check(old_result, new_result)

    new_result = jaxtorch.nn.functional.conv2d(x, w, padding='valid', dilation=2)
    old_result = torch.nn.functional.conv2d(x_torch, w_torch, padding='valid', dilation=2)
    check(old_result, new_result)

    w = jax.random.normal(key=rng.split(), shape=[4, 2, 3, 3])
    w_torch = totorch(w)
    new_result = jaxtorch.nn.functional.conv2d(x, w, padding='valid', groups=2)
    old_result = torch.nn.functional.conv2d(x_torch, w_torch, padding='valid', groups=2)
    check(old_result, new_result)

@torch.no_grad()
def test_conv2d_init():
    rng = PRNG(jax.random.PRNGKey(0))
    new_mod = jaxtorch.nn.Conv2d(256, 512, 3)
    new_weights = new_mod.init_weights(rng.split())
    old_mod = torch.nn.Conv2d(256, 512, 3)
    check(new_weights['weight'].std(), old_mod.weight.std(), 0.01)

@torch.no_grad()
def test_groupnorm():
    rng = PRNG(jax.random.PRNGKey(0))
    new = jaxtorch.nn.GroupNorm(8, 32)
    old = torch.nn.GroupNorm(8, 32)

    px = new.init_weights(rng.split())

    old.weight.data.copy_(totorch(px[new.weight.name]))
    old.bias.data.copy_(totorch(px[new.bias.name]))

    x = jax.random.normal(key=rng.split(), shape=[2, 32, 2])
    x_torch = totorch(x)

    cx = Context(px, rng.split())
    new_result = new(cx, x)
    old_result = old(x_torch)
    check(old_result, new_result)

@torch.no_grad()
def test_dropout():
    rng = PRNG(jax.random.PRNGKey(0))
    module = jaxtorch.nn.Dropout() # defaults to p=0.5

    px = module.init_weights(rng.split())

    x = jax.random.normal(key=rng.split(), shape=[1, 32])
    cx = Context(px, rng.split())

    out_train = module(cx.train_mode_(), x)
    assert (out_train != 0).sum() < 20
    out_eval = module(cx.eval_mode_(), x)
    assert (out_eval != 0).sum() == 32



"""Load and save jax state dicts and other objects as pytorch
checkpoint files.

"""

import jax
import jax.numpy as jnp
import jaxlib
import numpy as np
import torch

@torch.no_grad()
def load(f):
    """Converts torch.Tensor back to jax arrays after loading."""
    def from_torch(x):
        if isinstance(x, torch.Tensor):
            return jnp.asarray(x)
        return x
    torch_dict = torch.load(f, map_location='cpu')
    return jax.tree_util.tree_map(from_torch, torch_dict)

@torch.no_grad()
def save(obj, f):
    """Converts jax arrays (anything under jaxlib.xla_extension.DeviceArrayBase) to torch.Tensor before saving."""
    def to_torch(x):
        if isinstance(x, jaxlib.xla_extension.DeviceArrayBase):
            return torch.as_tensor(np.array(x))
        return x
    torch_dict = jax.tree_util.tree_map(to_torch, obj)
    torch.save(torch_dict, f)



import sys
sys.path = ['.'] + sys.path
import numpy as np
import jax
import jax.numpy as jnp
import jaxtorch

def check_near(x, y):
    assert (x - y).abs().max() < 1e-6, x-y

def test_upsample2x():
    image = jax.random.normal(jax.random.PRNGKey(0), (1,1,10,10))
    check_near(jax.image.resize(image, (1,1,20,20), method='linear'), jaxtorch.image.upsample2x(image, method='linear'))
    check_near(jax.image.resize(image, (1,1,20,20), method='cubic'), jaxtorch.image.upsample2x(image, method='cubic'))
    check_near(jax.image.resize(image, (1,1,20,20), method='lanczos3'), jaxtorch.image.upsample2x(image, method='lanczos3'))

def test_downsample2x():
    image = jax.random.normal(jax.random.PRNGKey(0), (1,1,20,20))
    check_near(jax.image.resize(image, (1,1,10,10), method='linear'), jaxtorch.image.downsample2x(image, method='linear'))
    check_near(jax.image.resize(image, (1,1,10,10), method='cubic'), jaxtorch.image.downsample2x(image, method='cubic'))
    check_near(jax.image.resize(image, (1,1,10,10), method='lanczos3'), jaxtorch.image.downsample2x(image, method='lanczos3'))



import jax
import jax.numpy as jnp
import jaxtorch
from jaxtorch.core import Module
from jaxtorch import init

class Upsample2d(Module):
    """Upsample image or features to 2x. Input: (n, c, h, w)."""
    def __init__(self, method='linear'):
        self.method = method
    def forward(self, cx, x):
        return jaxtorch.image.upsample2x(x, method=self.method)

class Downsample2d(Module):
    """Downsample image or features to 1/2x. Input: (n, c, h, w)."""
    def __init__(self, method='linear'):
        self.method = method
    def forward(self, cx, x):
        return jaxtorch.image.downsample2x(x, method=self.method)



import math
import jax
import jax.numpy as jnp
import numpy as np
from jaxtorch import core

def zeros(*shape):
    shape = jax.core.canonicalize_shape(shape)
    return core.Param(shape, lambda key: jnp.zeros(shape))

def ones(*shape):
    shape = jax.core.canonicalize_shape(shape)
    return core.Param(shape, lambda key: jnp.ones(shape))

def normal(*shape, stddev=1.0):
    shape = jax.core.canonicalize_shape(shape)
    return core.Param(shape, lambda key: stddev * jax.random.normal(key, shape))

def const(tensor):
    shape = jax.core.canonicalize_shape(tensor.shape)
    return core.Param(shape, lambda key: tensor)

def glorot_normal(*shape):
    shape = jax.core.canonicalize_shape(shape)
    fan_out = shape[0] * np.prod(shape[2:])
    fan_in = shape[1] * np.prod(shape[2:])
    stddev = np.sqrt(2.0 / (fan_in + fan_out))
    return core.Param(shape, lambda key: stddev * jax.random.normal(key, shape))

def uniform(*shape, min=-1.0, max=1.0):
    shape = jax.core.canonicalize_shape(shape)
    return core.Param(shape, lambda key: jax.random.uniform(key, shape, minval=min, maxval=max))

def kaiming_uniform(*shape, a=0):
    shape = jax.core.canonicalize_shape(shape)
    fan_in = np.prod(shape[1:])
    gain = math.sqrt(2.0 / (1 + a ** 2))
    bound = gain * math.sqrt(3.0 / fan_in)
    return uniform(*shape, min=-bound, max=bound)



"""Implement a bunch of convenience methods for jax arrays.

"""

import sys
import jax
import jax.numpy as jnp
import jaxlib
import numpy as np
import functools
from einops import rearrange, reduce, repeat

def register(**kwargs):
    for (attr, fun) in kwargs.items():
        if hasattr(jnp.zeros([]), attr):
            print(f'Not monkeypatching DeviceArray and Tracer with `{attr}`, because that method is already implemented.', file=sys.stderr)
            continue
        setattr(jaxlib.xla_extension.DeviceArrayBase, attr, fun)
        setattr(jax.interpreters.xla.DeviceArray, attr, fun)
        setattr(jax.core.Tracer, attr, fun)

def broadcast_to(arr, shape):
  shape = (shape,) if jnp.ndim(shape) == 0 else shape
  shape = jax.core.canonicalize_shape(shape)  # check that shape is concrete
  arr_shape = jax.core.canonicalize_shape(arr.shape)
  if arr_shape == shape:
      return arr
  nlead = len(shape) - len(arr_shape)
  shape_tail = shape[nlead:]
  compatible = all(arr_d in (1, shape_d) for (arr_d, shape_d) in zip(arr_shape, shape_tail))
  if nlead < 0 or not compatible:
      raise ValueError(f"Incompatible shapes for broadcasting: {arr_shape} and requested shape {shape}")
  diff = tuple(i for i, (arr_d, shape_d) in enumerate(zip(arr_shape, shape_tail)) if arr_d != shape_d)
  kept_dims = tuple(nlead + i for i in range(len(arr_shape)) if i not in diff)
  return jax.lax.broadcast_in_dim(jnp.squeeze(arr, diff), shape, kept_dims)

register(
    square = lambda arr: arr**2,
    sqrt = jnp.sqrt,
    abs = jnp.abs,
    add = lambda a, b: a + b,
    sub = lambda a, b: a - b,
    div = lambda a, b: a / b,
    mul = lambda a, b: a * b,
    sin = jnp.sin,
    cos = jnp.cos,
    arcsin = jnp.arcsin,
    arccos = jnp.arccos,
    log = jnp.log,
    exp = jnp.exp,
    expm1 = jnp.expm1,
    log1p = jnp.log1p,
    neg = lambda x: -x,
    clamp = lambda a, minval=None, maxval=None: jnp.clip(a, a_min=minval, a_max=maxval),
    unsqueeze = lambda arr, axis=0: jnp.expand_dims(arr, axis),
    rearrange = rearrange,
    reduce = reduce,
    broadcast_to = broadcast_to,
    isfinite = jnp.isfinite,
    isnan = jnp.isnan,
)



import jax
import jax.numpy as jnp

def conv2d(input, weight, bias=None, stride=1, padding=0, dilation=1, groups=1):
    (B, C1, iH, iW) = input.shape
    (C2, _, kH, kW) = weight.shape
    assert weight.shape[1] == C1//groups

    if type(padding) is int:
        padding = ((padding, padding), (padding, padding))
    elif type(padding) is str:
        padding = padding.upper()
    elif len(padding) == 2:
        padding = ((padding[0], padding[0]), (padding[1], padding[1]))

    if dilation > 1:
        dilation = (dilation, dilation)
    else:
        dilation = None

    output = jax.lax.conv_general_dilated(lhs=input,
                                          rhs=weight,
                                          window_strides=[stride, stride],
                                          padding=padding,
                                          lhs_dilation=None,
                                          rhs_dilation=dilation,
                                          dimension_numbers=('NCHW', 'OIHW', 'NCHW'),
                                          feature_group_count=groups)
    if bias is not None:
        output = output + bias.reshape(bias.shape[0], 1, 1)
    return output


def conv1d(input, weight, bias=None, stride=1, padding=0, dilation=1, groups=1):
    (B, C1, iH) = input.shape
    (C2, _, kH) = weight.shape
    assert weight.shape[1] == C1//groups

    if type(padding) is int:
        padding = [(padding, padding)]
    elif type(padding) is str:
        padding = padding.upper()
    elif len(padding) == 1:
        padding = ((padding[0], padding[0]),)

    if dilation > 1:
        dilation = (dilation,)
    else:
        dilation = None

    output = jax.lax.conv_general_dilated(lhs=input,
                                          rhs=weight,
                                          window_strides=[stride],
                                          padding=padding,
                                          lhs_dilation=None,
                                          rhs_dilation=dilation,
                                          dimension_numbers=('NCH', 'OIH', 'NCH'),
                                          feature_group_count=groups)
    if bias is not None:
        output = output + bias.reshape(bias.shape[0], 1)
    return output

def normalize(input, p=2.0, dim=1, eps=1e-12):
    if p != 2.0:
        raise NotImplementedError('only p=2.0 implemented so far')
    mag = input.square().sum(axis=dim, keepdims=True).sqrt()
    return input / mag.clamp(minval=eps)


import jax
import jax.numpy as jnp
import jaxlib
import numpy as np
import functools
import jaxtorch.monkeypatches
import sys
from dataclasses import dataclass

def _addindent(s_, numSpaces):
    s = s_.split('\n')
    # don't do anything for single-line stuff
    if len(s) == 1:
        return s_
    first = s.pop(0)
    s = [(numSpaces * ' ') + line for line in s]
    s = '\n'.join(s)
    s = first + '\n' + s
    return s

class Param(object):
    """Represents a parameter of a Module, and specifies its shape and initialization."""
    def __init__(self, shape, initializer):
        self.shape = shape
        self.initializer = initializer
        self.name = None

    def __repr__(self):
        if self.name is not None:
            return f'<Param at {self.name}>'
        else:
            return super().__repr__()

class PRNG(object):
    """Just a stateful wrapper for a jax.random.PRNGKey."""
    def __init__(self, key):
        self.key = key
    def split(self, n=None):
        (self.key, subkey) = jax.random.split(self.key)
        if n is not None:
            subkey = jax.random.split(subkey, n)
        return subkey

class ContextRandom(object):
    """Lives inside a Context and provides convenience functions for
random number generation that use the Context's stateful PRNG.

    """
    def __init__(self, rng):
        self.rng = rng

    def _wrap(f):
        @functools.wraps(f)
        def wrapped(self, *args, **kwargs):
            return f(self.rng.split(), *args, **kwargs)
        return wrapped

    bernoulli = _wrap(jax.random.bernoulli)
    beta = _wrap(jax.random.beta)
    categorical = _wrap(jax.random.categorical)
    cauchy = _wrap(jax.random.cauchy)
    choice = _wrap(jax.random.choice)
    dirichlet = _wrap(jax.random.dirichlet)
    double_sided_maxwell = _wrap(jax.random.double_sided_maxwell)
    exponential = _wrap(jax.random.exponential)
    gamma = _wrap(jax.random.gamma)
    gumbel = _wrap(jax.random.gumbel)
    laplace = _wrap(jax.random.laplace)
    logistic = _wrap(jax.random.logistic)
    maxwell = _wrap(jax.random.maxwell)
    multivariate_normal = _wrap(jax.random.multivariate_normal)
    normal = _wrap(jax.random.normal)
    pareto = _wrap(jax.random.pareto)
    permutation = _wrap(jax.random.permutation)
    poisson = _wrap(jax.random.poisson)
    rademacher = _wrap(jax.random.rademacher)
    randint = _wrap(jax.random.randint)
    shuffle = _wrap(jax.random.shuffle)
    t = _wrap(jax.random.t)
    truncated_normal = _wrap(jax.random.truncated_normal)
    uniform = _wrap(jax.random.uniform)
    weibull_min = _wrap(jax.random.weibull_min)

class Context(object):
    """Wraps a parameter dictionary and a PRNG."""
    def __init__(self, px, key, mode='train', tmp=None):
        self.px = px
        self.rng = PRNG(key)
        self.random = ContextRandom(self.rng)
        self.mode = mode
        self.tmp = dict(tmp or {})

    def train_mode_(self):
        self.mode = 'train'
        return self

    def eval_mode_(self):
        self.mode = 'eval'
        return self

    def __getitem__(self, par):
        if isinstance(par, Param):
            return self.px[par.name]
        elif isinstance(par, str):
            return self.px[par]
        else:
            raise TypeError('Expected a Param for indexing into Context')

    def __setitem__(self, par, value):
        if isinstance(par, Param):
            self.px[par.name] = value
        elif isinstance(par, str):
            self.px[par] = value
        else:
            raise TypeError('Expected a Param for indexing into Context')

    def freeze(self):
        return FrozenContext(self.px, self.rng.key, self.mode, self.tmp)

@jax.tree_util.register_pytree_node_class
@dataclass
class FrozenContext(object):
    params: dict
    key: jax.random.PRNGKey
    mode: str
    tmp: dict

    def thaw(self):
        return Context(self.params, self.key, self.mode, self.tmp)

    def thaw_into(self, cx):
        cx.rng.key = self.key
        cx.px.update(self.params)
        cx.tmp = dict(self.tmp)
        return cx

    def tree_flatten(self):
        return (self.params, self.key, self.tmp), (self.mode,)

    @staticmethod
    def tree_unflatten(static, dynamic):
        (params, key, tmp) = dynamic
        (mode,) = static
        return FrozenContext(params, key, mode, tmp)

def transform_with_cx(*transforms):
    def tr(func):
        # Inner wrapped function: pure functional, inputs and outputs frozen context
        def inner(frozen_cx, *args, **kwargs):
            cx = frozen_cx.thaw()
            result = func(cx, *args, **kwargs)
            return cx.freeze(), result
        # Transform pure function
        for t in reversed(transforms):
            inner = t(inner)
        # Wrap the transformed pure function to accept Context
        def outer(cx, *args, **kwargs):
            frozen_cx = cx.freeze()
            frozen_cx, result = inner(frozen_cx, *args, **kwargs)
            frozen_cx.thaw_into(cx)
            return result
        return outer
    return tr

class Module(object):
    name = None

    def __call__(self, cx: Context, *args, **kwargs):
        return self.forward(cx, *args, **kwargs)

    def forward(self, cx: Context, *args, **kwargs):
        """Implements the forward pass. Must take Context as the first argument."""
        raise NotImplementedError

    def self_named_modules(self):
        """Yields a sequence of (str, Module) for direct children of this
           module. May be overridden.

        """
        for (name, val) in self.__dict__.items():
            if isinstance(val, Module):
                yield (name, val)

    def self_named_parameters(self):
        """Yields a sequence of (str, Param) for direct children of this
           module. May be overridden.

        """
        for (name, val) in self.__dict__.items():
            if isinstance(val, Param):
                yield (name, val)

    def post_init_weights(self, cx):
        pass

    def self_init_weights(self, cx):
        """Initializes weights for this network's parameters. May be overriden
           for custom initialization. Child modules are initialized
           before parents.

        """
        for (name, mod) in self.self_named_modules():
            mod.self_init_weights(cx)
        for (name, par) in self.self_named_parameters():
            if par.initializer is not None:
                cx[par] = par.initializer(cx.rng.split())
        self.post_init_weights(cx)

    def name_everything_(self):
        for (name, par) in self.named_parameters():
            par.name = name
        for (name, mod) in self.named_modules():
            mod.name = name

    def init_weights(self, key):
        """Attaches names to parameters and returns initialized dict of
           parameters by name.

        """
        self.name_everything_()
        cx = Context({}, key)
        self.self_init_weights(cx)
        return cx.px

    def gen_named_modules(self):
        "Yields (str, Module) for all descendants of this module."
        for (name, val) in self.self_named_modules():
            yield (name, val)
            for (k, v) in val.gen_named_modules():
                yield (name+'.'+k, v)

    def gen_named_parameters(self):
        "Yields (str, Param) for this module and all descendants."
        for (name, par) in self.self_named_parameters():
            yield (name, par)

        for (name, mod) in self.self_named_modules():
            for (k, v) in mod.gen_named_parameters():
                yield (name+'.'+k, v)

    def named_parameters(self):
        return list(self.gen_named_parameters())

    def named_modules(self):
        return list(self.gen_named_modules())

    def modules(self):
        return [m for (k, m) in self.gen_named_modules()]

    def parameters(self):
        return [p for (k, p) in self.gen_named_parameters()]

    def state_dict(self, px):
        return {name:px[par.name] for (name, par) in self.gen_named_parameters()}

    def load_state_dict(self, px, state, strict=True):
        """Load a previously saved state_dict into px. Returns px."""
        for (k, p) in self.gen_named_parameters():
            if k not in state:
                if strict:
                    raise ValueError(f'Not loading missing parameter: {k}')
                else:
                    print(f'Not loading missing parameter: {k}', file=sys.stderr)
                    continue

            if px[p.name].shape != state[k].shape:
                msg = f'Not loading parameter from incompatible shape: {k} ({px[p.name].shape} vs {state[k].shape})'
                if strict:
                    raise ValueError(msg)
                else:
                    print(msg, file=sys.stderr)
                    continue

            px[p.name] = jax.numpy.asarray(state[k])
        return px

    def _get_name(self):
        return self.__class__.__name__

    def extra_repr(self) -> str:
        r"""Set the extra representation of the module

        To print customized extra information, you should re-implement
        this method in your own modules. Both single-line and multi-line
        strings are acceptable.
        """
        return ''

    def __repr__(self):
        # We treat the extra repr like the sub-module, one item per line
        extra_lines = []
        extra_repr = self.extra_repr()
        # empty string will be split into list ['']
        if extra_repr:
            extra_lines = extra_repr.split('\n')
        child_lines = []
        for key, module in self.self_named_modules():
            mod_str = repr(module)
            mod_str = _addindent(mod_str, 2)
            child_lines.append('(' + key + '): ' + mod_str)
        lines = extra_lines + child_lines

        main_str = self._get_name() + '('
        if lines:
            # simple one-liner info, which most builtin Modules will use
            if len(extra_lines) == 1 and not child_lines:
                main_str += extra_lines[0]
            else:
                main_str += '\n  ' + '\n  '.join(lines) + '\n'

        main_str += ')'
        return main_str



import math
import jax
import jax.numpy as jnp
import jaxtorch
import numbers
from jaxtorch.core import Module, PRNG, Context
from jaxtorch import init

class Identity(Module):
    def forward(self, cx, x):
        return x


class ModuleList(Module):
    def __init__(self, *items):
        self.items = []
        for item in items:
            if isinstance(item, Module):
                self.items.append(item)
            elif isinstance(item, (list, tuple)):
                self.items.extend(item)
            else:
                raise ValueError("Expected module or sequence to ModuleList()")

    def __iter__(self):
        return iter(self.items)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return type(self)(self.items[index])
        elif isinstance(index, int):
            return self.items[index]
        else:
            raise TypeError("Invalid argument type.")

    def append(self, mod):
        self.items.append(mod)

    def forward(self, cx, x):
        raise NotImplementedError

    def self_named_modules(self):
        for (i, m) in enumerate(self.items):
            yield (f'{i}', m)


class Sequential(ModuleList):
    def forward(self, cx, x):
        for module in self:
            x = module(cx, x)
        return x


class Linear(Module):
    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = init.glorot_normal(out_features, in_features)
        if bias:
            self.bias = init.zeros(out_features)
        else:
            self.bias = None

    def forward(self, cx, x):
        y = x @ jnp.transpose(cx[self.weight])
        if self.bias:
            y = y + cx[self.bias]
        return y

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )


class Embedding(Module):
    def __init__(self, num_embeddings: int, embedding_dim: int):
        super().__init__()
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        self.weight = init.normal(num_embeddings, embedding_dim)

    def forward(self, cx, x):
        return cx[self.weight][x]

    def extra_repr(self) -> str:
        s = '{num_embeddings}, {embedding_dim}'
        # if self.padding_idx is not None:
        #     s += ', padding_idx={padding_idx}'
        # if self.max_norm is not None:
        #     s += ', max_norm={max_norm}'
        # if self.norm_type != 2:
        #     s += ', norm_type={norm_type}'
        # if self.scale_grad_by_freq is not False:
        #     s += ', scale_grad_by_freq={scale_grad_by_freq}'
        # if self.sparse is not False:
        #     s += ', sparse=True'
        return s.format(**self.__dict__)



class Tanh(Module):
    def forward(self, cx, x):
        return jnp.tanh(x)


class Dropout(Module):
    def __init__(self, p=0.5):
        self.rate = p

    def forward(self, cx, x):
        if cx.mode == 'eval':
            return x
        mask = cx.random.bernoulli(1.0 - self.rate, shape=x.shape)
        return x * mask / (1.0 - self.rate)

class Dropout2d(Module):
    def __init__(self, p=0.5):
        self.rate = p

    def forward(self, cx, x):
        if cx.mode == 'eval':
            return x
        drop_shape = x.shape[:2] + (1,) * len(x.shape[2:])
        mask = cx.random.bernoulli(1.0 - self.rate, shape=drop_shape)
        return x * mask / (1.0 - self.rate)

class Sigmoid(Module):
    def forward(self, cx, x):
        return jax.nn.sigmoid(x)

class GELU(Module):
    def forward(self, cx, x):
        return jax.nn.gelu(x)

class ReLU(Module):
    def forward(self, cx, x):
        return jax.nn.relu(x)

class LeakyReLU(Module):
    def __init__(self, negative_slope=0.01):
        self.negative_slope = negative_slope

    def forward(self, cx, x):
        return jax.nn.leaky_relu(x, self.negative_slope)


class LayerNorm(Module):
    def __init__(self, normalized_shape, eps=1e-05, elementwise_affine=True):
        if isinstance(normalized_shape, numbers.Integral):
            normalized_shape = (normalized_shape,)
        self.normalized_shape = tuple(normalized_shape)
        self.eps = eps
        self.elementwise_affine = elementwise_affine
        if elementwise_affine:
            self.weight = init.ones(*normalized_shape)
            self.bias = init.zeros(*normalized_shape)
        else:
            self.weight = None
            self.bias = None
        self.axes = tuple(-i for i in range(1, len(normalized_shape)+1))

    def forward(self, cx, x):
        mu = x.mean(axis=self.axes, keepdims=True)
        sigma = jnp.sqrt((x - mu).square().mean(axis=self.axes, keepdims=True) + self.eps)
        normed = (x - mu) / sigma
        return cx[self.weight] * normed + cx[self.bias]


class Conv1d(Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, zero_init=False):
        assert in_channels % groups == 0
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.weight = init.kaiming_uniform(out_channels, in_channels//groups, kernel_size, a=math.sqrt(5.0))
        if zero_init:
            self.weight = init.zeros(out_channels, in_channels//groups, kernel_size)
        self.use_bias = bias
        if self.use_bias:
            self.bias = init.zeros(out_channels)
        else:
            self.bias = None

    def forward(self, cx, x):
        return jaxtorch.nn.functional.conv1d(x, cx[self.weight], cx[self.bias] if self.use_bias else None,
                                             stride=self.stride,
                                             padding=self.padding,
                                             dilation=self.dilation,
                                             groups=self.groups)


class Conv2d(Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, zero_init=False):
        assert in_channels % groups == 0
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.weight = init.kaiming_uniform(out_channels, in_channels//groups, kernel_size, kernel_size, a=math.sqrt(5.0))
        if zero_init:
            self.weight = init.zeros(out_channels, in_channels//groups, kernel_size, kernel_size)
        self.use_bias = bias
        if self.use_bias:
            if zero_init:
                self.bias = init.zeros(out_channels)
            else:
                fan_in = in_channels//groups * kernel_size**2
                bound = 1 / math.sqrt(fan_in)
                self.bias = init.uniform(out_channels, min=-bound, max=bound)
        else:
            self.bias = None

    def forward(self, cx, x):
        return jaxtorch.nn.functional.conv2d(x, cx[self.weight], cx[self.bias] if self.use_bias else None,
                                             stride=self.stride,
                                             padding=self.padding,
                                             dilation=self.dilation,
                                             groups=self.groups)


class SiLU(Module):
    def forward(self, cx, x):
        return jax.nn.silu(x)

class GroupNorm(Module):
    def __init__(self, num_groups, num_channels, eps=1e-05, affine=True):
        self.num_groups = num_groups
        self.num_channels = num_channels
        assert self.num_channels % self.num_groups == 0
        self.eps = eps
        self.affine = affine
        if self.affine:
            self.weight = init.ones(num_channels)
            self.bias = init.zeros(num_channels)
        else:
            self.weight = None
            self.bias = None

    def forward(self, cx, x):
        B, C, *rest = x.shape
        assert C == self.num_channels
        x = x.reshape([B, self.num_groups, C//self.num_groups, *rest])
        mu = x.mean(axis=tuple(range(2,len(x.shape))), keepdims=True)
        var = x.var(axis=tuple(range(2,len(x.shape))), keepdims=True)
        y = (x - mu) / jnp.sqrt(var + self.eps)
        y = y.reshape([B, C, *rest])
        if self.affine:
            broadcast_shape = [self.num_channels] + [1] * len(rest)
            weight = cx[self.weight].reshape(broadcast_shape)
            bias = cx[self.bias].reshape(broadcast_shape)
            y = y * weight + bias
        return y

class PixelUnshuffle(Module):
    def __init__(self, downscale_factor):
        self.downscale_factor = downscale_factor
    def forward(self, cx, x):
        return x.rearrange('... c (h r) (w s) -> ... (c r s) h w', r = self.downscale_factor, s = self.downscale_factor)

class PixelShuffle(Module):
    def __init__(self, upscale_factor):
        self.upscale_factor = upscale_factor
    def forward(self, cx, x):
        return x.rearrange('... (c r s) h w -> ... c (h r) (w s)', r = self.upscale_factor, s = self.upscale_factor)


from setuptools import find_packages, setup
import os
import pathlib
import pkg_resources

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

with open('jaxtorch/_version.py') as fp:
    exec(fp.read())

# This call to setup() does all the work
setup(
    name="jaxtorch",
    version=__version__,
    description="A jax based nn library",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/nshepperd/jaxtorch",
    author="Emily Shepperd",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(include=['jaxtorch*']),
    include_package_data=True,
    install_requires=[
        str(r)
        for r in pkg_resources.parse_requirements(
            open(os.path.join(os.path.dirname(__file__), "requirements.txt"))
        )
    ],
    extras_require={'dev': []},
)
